# This file is part of SSHare.
#
# SSHare is free software: you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# SSHare is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# SSHare. If not, see <https://www.gnu.org/licenses/>.

import argparse
import getpass
import importlib
import importlib.util
import os
import os.path
import pyclip
import time
import tomllib
import subprocess
import sys
from pathlib import Path
from version import version

class Config:
    def __init__(self, config_directory):
        _config = self._load_from_file(config_directory / "config.toml")

        self.source_directory = _config.get("source_directory")

        host = _config.get("host")
        if host == None:
            print("Error: 'host' cannot be 'None'")
            sys.exit(1)
        self.host_protocol = host.get("protocol")
        self.host_name     = host.get("name")
        self.host_port     = host.get("port")
        self.host_path     = host.get("path")
        if self.host_protocol == None:
            self.host_protocol = "https"
        if self.host_name == None:
            print("Error: 'host.name' cannot be 'None'")
            sys.exit(1)
        if self.host_port == None:
            self.host_port = ""
        else:
            self.host_port = f":{self.host_port}"
        if self.host_path == None:
            self.host_path = ""
        
        ssh = _config.get("ssh")
        if ssh == None:
            print("Error: 'ssh' cannot be 'None'")
            sys.exit(1)
        self.ssh_port = ssh.get("port")
        self.ssh_user = ssh.get("user")
        self.ssh_path = ssh.get("path")
        if self.ssh_port == None:
            self.ssh_port = 22
        if self.ssh_user == None:
            self.ssh_user = getpass.getuser()
        if self.ssh_path == None:
            print("Error: 'ssh.path' cannot be 'None'")
            sys.exit(1)

    def _load_from_file(self, config_path):
        with open(config_path, mode="rb") as file:
            return tomllib.load(file)


def rebase(number):
    if number == 0:
        return "0"
    rebased = ""
    while number != 0:
        digit = number % 62
        if digit < 10:
            rebased = chr(digit + 48) + rebased
        elif digit < 36:
            rebased = chr(digit + 87) + rebased
        else:
            rebased = chr(digit + 29) + rebased
        number = int(number / 62)
    return rebased


class MetaLogger:
    def __init__(self, loggers):
        def info(message):
            for logger in loggers:
                logger.info(message)

        def warn(message):
            for logger in loggers:
                logger.warn(message)

        def error(message):
            for logger in loggers:
                logger.error(message)

        self.info  = info
        self.warn  = warn
        self.error = error

class Plugin:
    def __init__(self, name, module):
        self.name   = name
        self.module = module

def main():
    config_directory = Path(os.environ.get("XDG_CONFIG_DIR") or f"{os.environ["HOME"]}/.config") / "sshare"
    logger = importlib.import_module("command_line_logger")
    modules = {
        "logger": [ Plugin("command_line_logger", logger) ],
        "data": [],
    }
    error = False
    sys.dont_write_bytecode = True
    for plugin in (config_directory / "plugins").iterdir():
        if plugin.is_file():
            module_spec = importlib.util.spec_from_file_location(
                plugin.stem,
                plugin.as_posix(),
            )
            module = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(module)

            modules_of_type = modules.get(module.plugin_type)
            if modules_of_type == None:
                logger.error(f"Error: Plugin '{plugin.stem}' has an invalid plugin type '{module.plugin_type}'")
                error = True
            else:
                modules_of_type.append(Plugin(plugin.stem, module))
                if hasattr(module, "init"):
                    module.init()
    if error:
        sys.exit(1)
    sys.dont_write_bytecode = False
    logger = MetaLogger([ logger.module for logger in modules["logger"] ])
    logger.info("Successfully loaded plugins")

    arguments = parse_arguments()

    config = Config(config_directory)

    contents = b''
    target_file_extension = ""
    if arguments.latest or arguments.file != None:
        file_path = ""
        if arguments.latest:
            if config.source_directory == "":
                print("Option 'latest' requires source directory to be specified")
                sys.exit(1)
            file_path = _latest(config.source_directory)
        else:
            file_path = arguments.file
        print(f"Uploading file '{file_path}'")

        with open(file_path, mode="rb") as file:
            contents = file.read()

        (_, target_file_extension) = os.path.splitext(file_path)
    elif arguments.paste:
        print("Uploading contents of clipboard")
        contents = pyclip.paste()
        target_file_extension = ".txt"
    else:
        print("Error: must specify one of -f FILE, -l, -p")
        sys.exit(1)
       
    target_id = rebase(time.time_ns())
    target_file_name = f"{target_id}{target_file_extension}"
    target_file = f"{config.ssh_path}/{target_file_name}"
    target_destination = f"{config.ssh_user}@{config.host_name}"
    print(f"Uploading to host: {target_destination}, port: {config.ssh_port}, file: {target_file}")
    process = subprocess.run([
            "ssh",
            f"-p {config.ssh_port}",
            target_destination,
            "-T",
            f"cat - > {target_file}"
        ],
        input = contents,
    )
    if process.returncode != 0:
        print("Error: failed to upload file")
        sys.exit(1)

    target_url = f"{config.host_protocol}://{config.host_name}{config.host_port}{config.host_path}/{target_file_name}"
    print(f"File available at '{target_url}'")
    if arguments.copy:
        pyclip.copy(target_url)
        print("URL copied to clipboard")

    sys.exit(0)


def parse_arguments():
    parser = argparse.ArgumentParser(
        prog = "SSHare",
        description = "Upload files to a server via ssh",
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s version {version}",
    )
    parser.add_argument(
        "-l",
        "--latest",
        action="store_const",
        const=True,
        help="Upload the latest image from the source directory",
    )
    parser.add_argument(
        "-p",
        "--paste",
        action="store_const",
        const=True,
        help="Upload the contents of the clipboard as a .txt file",
    )
    parser.add_argument(
        "-f",
        "--file",
        help="Upload a file",
    )
    parser.add_argument(
        "-c",
        "--copy",
        action="store_const",
        const=True,
        help="Copy the resultant URL to the clipboard",
    )
    arguments = parser.parse_args()

    return arguments


def _latest(directory, key=os.path.getmtime):
    files         = map(lambda file: f"{directory}/{file}", os.listdir(directory))
    selection     = next(files)
    selection_key = key(selection)
    for file in files:
        new_key = key(file)
        if new_key > selection_key:
            selection     = file
            selection_key = key
    return selection
